/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of siralina.
   
   siralina is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   siralina is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with siralina.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <libxml/parser.h>

#include <stdarg.h>

#include <GDD/arch_xml.h>
#include <GDD/ddg_xml.h>

#include <SIRA/reduce.h>
#include <RS/greedy.h>

#include <assert.h>

void usage(const char *app) {
  fprintf(stdout,"%s -a arch.xml -i ddg.xml [-t regtype #value]\n",app);
}

#define WARNING 0
#define ERROR 1

void warn_error(int status, const char *app, const char *fmt, ...) {
  va_list ap;
  if(status == WARNING) {
    fprintf(stderr,"warning: ");
  } else {
    fprintf(stderr,"error: ");
  }
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fprintf(stderr,"\n");
  usage(app);
  if(status != WARNING) {
    exit(status);
  }
}

int main(int argc, char *argv[]) {
  GDD_Arch *arch = NULL;
  SCEDA_Graph *g = NULL;

  int IImin_rec;

  SIRA_WeightMap *weight_map = SIRA_weight_map_create();

  xmlInitParser();

  int i;
  for(i = 0; i < argc; i++) {
    if(strcmp(argv[i],"-i") == 0) {
      i++;
      if(i < argc) {
	if(g == NULL) {
	  char *filename = argv[i];
	  if(arch == NULL) {
	    fprintf(stderr,"warning: no arch file given\n");
	    fprintf(stderr,"attempting to load default arch file\n");
	    arch = GDD_arch_of_xml_file("default_arch.xml");
	    assert(arch != NULL);
	  }
	  fprintf(stderr,"loading ddg file %s\n",filename);
	  g = GDD_of_xml_file(TRUE, arch, filename, &IImin_rec);
	  assert(g != NULL);
	} else {
	  warn_error( ERROR, argv[0], "ddg file already given");
	}
      } 
    } else if(strcmp(argv[i],"-a") == 0) {
      i++;
      if(i < argc) {
	if(arch == NULL) {
	  char *filename = argv[i];
	  fprintf(stderr,"loading arch file %s\n",filename);
	  arch = GDD_arch_of_xml_file(filename);
	  assert(arch != NULL);
	} else {
	  warn_error( ERROR, argv[0], "arch file already given");
	}
      }
    } else if(strcmp(argv[i],"-t") == 0) {
      i++;
      if(i < argc) {
	char *type = argv[i];
	i++;
	if(i < argc) {
	  int w = atoi(argv[i]);
	  if(w < 0) {
	    warn_error( ERROR, argv[0], "invalid weight %d", w);
	  } else {
	    if(SIRA_weight_map_set_weight_of_type(weight_map, type, w) != 0) {
	      warn_error( ERROR, argv[0], "problem when setting weight of type %s", type);
	    }
	  }
	} else {
	  warn_error( ERROR, argv[0], "missing weight");
	}
      } else {
	warn_error( ERROR, argv[0], "missing register type");
      }
    } 
  }

  xmlCleanupParser();

  if(g == NULL) {
    warn_error( ERROR, argv[0], "no ddg file given");
  }

  {
    GDD_DAG *dag = GDD_dag_create(g);
    
    GDD_ArchTypesIterator all_types;
    GDD_arch_types_iterator_init(arch, &all_types);
    while(GDD_arch_types_iterator_has_next(&all_types)) {
      char *type = GDD_arch_types_iterator_next(&all_types);
      fprintf(stdout, "RS(%s)=%d\n", type, RS_ddag_estimate_saturation(dag, type, TRUE)); 
    }
    GDD_arch_types_iterator_cleanup(&all_types);
    
    GDD_dag_delete(dag);
  }

  SCEDA_List *added_edges;

  if(SIRA_weight_map_is_empty(weight_map)) {
    GDD_ArchTypesIterator all_types;
    GDD_arch_types_iterator_init(arch, &all_types);
    while(GDD_arch_types_iterator_has_next(&all_types)) {
      char *type = GDD_arch_types_iterator_next(&all_types);
      SIRA_weight_map_set_weight_of_type(weight_map, type, 1);
    }
    GDD_arch_types_iterator_cleanup(&all_types);
  }

  int II = SIRA_reduce(g, weight_map, &added_edges);

  if(II >= 0) {
    fprintf(stdout,"success\n");
    fprintf(stdout,"%d edges added\n", SCEDA_list_size(added_edges));
  } else {
    fprintf(stdout,"failure\n");
  }

  {
    GDD_DAG *dag = GDD_dag_create(g);
    
    GDD_ArchTypesIterator all_types;
    GDD_arch_types_iterator_init(arch, &all_types);
    while(GDD_arch_types_iterator_has_next(&all_types)) {
      char *type = GDD_arch_types_iterator_next(&all_types);
      fprintf(stdout, "RS(%s)=%d\n", type, RS_ddag_estimate_saturation(dag, type, TRUE)); 
    }
    GDD_arch_types_iterator_cleanup(&all_types);
    
    GDD_dag_delete(dag);
  }

  if(II >= 0) {
    GDD_delete_edges(g, added_edges);
    SCEDA_list_delete(added_edges);
  }

  SIRA_weight_map_delete(weight_map);

  SCEDA_graph_delete(g);
  GDD_arch_delete(arch);

  return 0;
}
