/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>

#include <SCEDA/graph_flow.h>
#include <SCEDA/queue.h>
#include <SCEDA/hashmap.h>

#include "solver.h"
#include "util.h"
#include "hungarian/hungarian.h"

/* Include string match and hash functions */
#include "stringutil.h"

#include "boxed.h"

#include "reuse.h"

// Define USE_ILP to use an Integer Linear Program solver (GLPK, LPSolve)
// otherwise use the network flows implementation

#ifdef USE_ILP
#warning "Using Integer Linear Programming implementation for SiraLina"
// define USE_LPSOLVE to use LPSolve
// otherwise use GLPK
//#define PROBEOBJFUN
//#define DEBUG

#define SIRA_solver_fixed_period_ilp SIRA_solver_fixed_period

#ifdef PROBEOBJFUN
int obj_fun;
#endif

#ifdef USE_LPSOLVE
#warning "ILP solver is LP Solve"
#include "lp_solve_stub.h"
#else
#warning "ILP solver is GLPK (GPL tainted binary)"
#include "glpk_stub.h"
#endif
#else
#warning "Using Network Flows implementation for SiraLina"
#define SIRA_solver_fixed_period_mcf SIRA_solver_fixed_period
#endif

#define MAX_NAME_SIZE 128

/** Functions to easily manipulate matrix as used by the Hungarian library */
/** Create a squared matrix of desired size */
static int **matrix_create(int size) {
  int **matrix = safe_calloc(size, sizeof(int *));
  int i;
  for(i = 0; i < size; i++) {
    matrix[i] = safe_calloc(size, sizeof(int));
  }
  return matrix;
}

/** Delete a squared matrix of given size */
static void matrix_delete(int size, int **matrix) {
  int i;
  for(i = 0; i < size; i++) {
    free(matrix[i]);
  }
  free(matrix);
}

#ifdef USE_ILP
/** The State data structure helps to keep track of correspondance
    between variables of the linear problem and vertices of the
    graph */
typedef struct {
  int idx; // for name generation
  int sigma; // sigma
  SCEDA_HashMap *sigma_k_map; // sigma of killers (indexed by types)
} State;

static State *state_create() {
  State *st = (State *)safe_malloc(sizeof(State));
  st->sigma_k_map = SCEDA_hashmap_create(NULL, (SCEDA_delete_fun)boxed_delete, (SCEDA_compare_fun)string_match, (SCEDA_hash_fun)string_hash);
  return st;
}

static void state_delete(State *x) {
  SCEDA_hashmap_delete(x->sigma_k_map);
  free(x);
}

/** Initialise the ILP solver 

    @param[in] g = G
    @param[in] weight_map = weight function (types -> boxed(int)) 
    @param[in] states = map from vertices to states

    initialise the states map and return the number of needed
    variables (number of columns)
*/
static int SIRA_solver_init(SCEDA_Graph *g, SIRA_WeightMap *weight_map, SCEDA_HashMap *states) {
  int nb = 0;
  int idx = 0;

  SCEDA_VerticesIterator vertices;
  SCEDA_vertices_iterator_init(g, &vertices);
  while(SCEDA_vertices_iterator_has_next(&vertices)) {
    SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);
    State *st_v = state_create();
    SCEDA_hashmap_put(states, v, st_v, NULL);
    st_v->idx = idx++;
    st_v->sigma = ++nb;
    
    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }
      if(GDD_vertex_is_value_of_type(v, type)) {
	SCEDA_hashmap_put(st_v->sigma_k_map, type, boxed_create(int, ++nb), NULL);
      }
    }
    SIRA_weight_map_iterator_cleanup(&weights);
  }
  SCEDA_vertices_iterator_cleanup(&vertices);

  return nb;
}

/** Solve the assignment problem

    @param[in] scheduling_prob = scheduling problem (solved)
    @param[in] vals = V_{R,t}
    @param[in] type = type t
    @param[in] ual_flag = true if UAL semantics
    @param[in] period = scheduling period
    @param[in] states = map V -> State
    @param[in] reuse_edges = list of reuse edges (modified)

    return the cost of the minimal assignment
 */
static int SIRA_solver_assignment_problem(ILP_SOLVER_TYPE *scheduling_prob, 
				     SCEDA_HashSet *vals, const char *type,
				     int ual_flag, int period, 
				     SCEDA_HashMap *states,
				     SCEDA_List *reuse_edges) {
  int n = SCEDA_hashset_size(vals);

  SCEDA_Vertex *values[n];

  {
    int i = 0;
    SCEDA_HashSetIterator vs;
    SCEDA_hashset_iterator_init(vals, &vs);
    while(SCEDA_hashset_iterator_has_next(&vs)) {
      SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&vs);
      values[i++] = v;
    }
    SCEDA_hashset_iterator_cleanup(&vs);
  }

  int **assignment_matrix = matrix_create(n);

  {
    int i;
    for(i = 0; i < n; i++) {
      SCEDA_Vertex *vi = values[i];
      State *st_vi = SCEDA_hashmap_get(states, vi);
      
      boxed(int) sigma_ki_col = SCEDA_hashmap_get(st_vi->sigma_k_map, type);
      
      int sigma_ki = ilp_solver_get_col_value(scheduling_prob, boxed_get(sigma_ki_col));
      int j;
      for(j = 0; j < n; j++) {
	SCEDA_Vertex *vj = values[j];
	State *st_vj = SCEDA_hashmap_get(states, vj);
	int sigma_j = ilp_solver_get_col_value(scheduling_prob, st_vj->sigma);
	
	int mu_ij = sigma_ki - sigma_j;
	if(ual_flag) {
	  mu_ij++;
	} else {
	  mu_ij -= GDD_vertex_delta_w(vj, type);
	}
	
	mu_ij += period - 1;
	if(mu_ij < 0) {
	  mu_ij = 0;
	}
	mu_ij /= period;
	
	assignment_matrix[i][j] = mu_ij;
      }
    }
  }

  hungarian_problem_t assignment_problem;

  int size = hungarian_init(&assignment_problem, assignment_matrix, n, n, HUNGARIAN_MODE_MINIMIZE_COST);
  assert(size == n);

  hungarian_solve(&assignment_problem);   
  
  int cost = 0;
  {
    int i;
    for(i = 0; i < n; i++) {
      int j;
      for(j = 0; j < n; j++) {
	if(assignment_problem.assignment[i][j] == HUNGARIAN_ASSIGNED) {
	  int mu_ij = assignment_matrix[i][j];
	  cost += mu_ij;
	  SIRA_ReuseEdge *e = SIRA_reuse_edge_create(values[i], values[j], mu_ij);
	  safe_call(SCEDA_list_add(reuse_edges, e));
	}
      }
    }
  }

  hungarian_free(&assignment_problem);

  matrix_delete(n, assignment_matrix);

  return cost;
}

/** Siralina solver (use ILP solver)

    @param[in] loop = Siraloop DDG
    @param[in] settings = solver settings
    @param[in] period = scheduling period (II)
    @param[in] reuse_decisions = map (type -> reuse edges list) 

    return 0 in case of success, -1 otherwise */
static int SIRA_solver_fixed_period_ilp(SIRA_Loop *loop, 
					SIRA_Param *settings,
					SIRA_WeightMap *weight_map,
					int period,
					SCEDA_HashMap *reuse_decisions) {
  int ual_flag = settings->ual_flag;
  int bounded_flag = settings->bounded_flag;

  int result = -1; // error

  int Lmax = 0;
  if(bounded_flag) {
    Lmax = SIRA_loop_get_L(loop, period);
  }

  SCEDA_Graph *g = SIRA_loop_ddg(loop);

  char namebuf[MAX_NAME_SIZE];

  ilp_solver_init();

  // compute number of columns
  SCEDA_HashMap *states = SCEDA_vertex_map_create((SCEDA_delete_fun)state_delete);

  int cols = SIRA_solver_init(g, weight_map, states);

  ILP_SOLVER_TYPE *scheduling_prob = ilp_solver_create("Scheduling Problem", TRUE, cols);
  if(scheduling_prob == NULL) {
    fprintf(stderr,"SIRA: error while creating linear problem\n");
    SCEDA_hashmap_delete(states);
    ilp_solver_cleanup();
    return -1;
  }

  // set variables names, bounds & objective function
  {
    SCEDA_HashMapIterator vertices;
    SCEDA_hashmap_iterator_init(states, &vertices);
    while(SCEDA_hashmap_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v;
      State *st_v = SCEDA_hashmap_iterator_next(&vertices, &v);

      //      fprintf(stdout,"%s\n", GDD_vertex_mnemonic(v));

      // set name
      snprintf(namebuf, MAX_NAME_SIZE, "sigma(%d)", st_v->idx);
      ilp_solver_set_col_name(scheduling_prob, st_v->sigma, namebuf);

      // set bounds
      if(bounded_flag) {
	int s_min = SIRA_loop_sigma_min(loop, v);
	int s_max = SIRA_loop_sigma_max(loop, v, Lmax);
	ilp_solver_set_col_bounds(scheduling_prob, st_v->sigma, ILP_SOLVER_DB, s_min, s_max);
      } else {
	ilp_solver_set_col_bounds(scheduling_prob, st_v->sigma, ILP_SOLVER_LO, 0, 0); // sigma_i >= 0
      }

      // set objective function coef
      {
	double coef = 0;
	
	SCEDA_HashMapIterator sigma_ks;
	SCEDA_hashmap_iterator_init(st_v->sigma_k_map, &sigma_ks);
	while(SCEDA_hashmap_iterator_has_next(&sigma_ks)) {
	  char *type;
	  boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&sigma_ks, &type);
	  
	  // set name
	  snprintf(namebuf, MAX_NAME_SIZE, "sigma_k(%d,%s)", st_v->idx, type);
	  ilp_solver_set_col_name(scheduling_prob, boxed_get(sigma_k), namebuf);
	  
	  // set bounds
	  if(bounded_flag) {
	    int k_min = SIRA_loop_k_min(loop, v, type, period);
	    int k_max = SIRA_loop_k_max(loop, v, Lmax, type, period);
	    ilp_solver_set_col_bounds(scheduling_prob, boxed_get(sigma_k), ILP_SOLVER_DB, k_min, k_max); // sigmak_i >= 0
	  } else {
	    ilp_solver_set_col_bounds(scheduling_prob, boxed_get(sigma_k), ILP_SOLVER_LO, 0, 0); // sigmak_i >= 0
	  }
	  
	  // set objective function coef
	  int alpha = SIRA_weight_map_get_weight_of_type(weight_map, type);
	  
	  ilp_solver_set_obj_coef(scheduling_prob, boxed_get(sigma_k), alpha);
	  
	  coef += alpha;
	}
	SCEDA_hashmap_iterator_cleanup(&sigma_ks);
	
	// set objective function coef
	ilp_solver_set_obj_coef(scheduling_prob, st_v->sigma, -coef);
      }
    }
    SCEDA_hashmap_iterator_cleanup(&vertices);
  }

  // set linear constraints
  {
    int ind[1+2];
    ILP_SOLVER_COEF_TYPE coef[1+2];

    SCEDA_EdgesIterator edges;
    SCEDA_edges_iterator_init(g, &edges);
    while(SCEDA_edges_iterator_has_next(&edges)) {
      SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);

      SCEDA_Vertex *v_s = SCEDA_edge_source(e);
      SCEDA_Vertex *v_t = SCEDA_edge_target(e);
      
      State *st_s = SCEDA_hashmap_get(states, v_s);
      State *st_t = SCEDA_hashmap_get(states, v_t);

      snprintf(namebuf, MAX_NAME_SIZE, "sigma(%d)-sigma(%d)", st_t->idx, st_s->idx);
      int lowb = GDD_edge_latency(e) - period * GDD_edge_distance(e);
      if(v_s == v_t) {
	ilp_solver_add_row(scheduling_prob, namebuf, 0, ind, coef, ILP_SOLVER_LO, lowb);
      } else {
	ind[1] = st_s->sigma;
	coef[1] = -1;
	ind[2] = st_t->sigma;
	coef[2] = 1;
	ilp_solver_add_row(scheduling_prob, namebuf, 2, ind, coef, ILP_SOLVER_LO, lowb);
      }
    }
    SCEDA_edges_iterator_cleanup(&edges);

    SCEDA_HashMapIterator vertices;
    SCEDA_hashmap_iterator_init(states, &vertices);
    while(SCEDA_hashmap_iterator_has_next(&vertices)) {
      SCEDA_Vertex *u;
      State *st_u = SCEDA_hashmap_iterator_next(&vertices, &u);

      SCEDA_HashMapIterator sigma_ks;
      SCEDA_hashmap_iterator_init(st_u->sigma_k_map, &sigma_ks);
      while(SCEDA_hashmap_iterator_has_next(&sigma_ks)) {
	char *type;
	boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&sigma_ks, &type);
	
	SCEDA_HashSet *cons_u = GDD_consumers_of_type(u, type);
	if(SCEDA_hashset_is_empty(cons_u)) {
	  snprintf(namebuf, MAX_NAME_SIZE, "sigma_k(%d,%s)", st_u->idx, type);
	  ind[1] = boxed_get(sigma_k);
	  coef[1] = 1;
	  ind[2] = st_u->sigma;
	  coef[2] = -1;
	  int lowb = 1;
	  ilp_solver_add_row(scheduling_prob, namebuf, 2, ind, coef, ILP_SOLVER_LO, lowb);
	} else {
	  SCEDA_HashSetIterator consumers;
	  SCEDA_hashset_iterator_init(cons_u, &consumers);
	  while(SCEDA_hashset_iterator_has_next(&consumers)) {
	    SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&consumers);
	    State *st_v = SCEDA_hashmap_get(states, v);

	    snprintf(namebuf, MAX_NAME_SIZE, "sigma_k(%d,%s)-sigma(%d)", st_u->idx, type, st_v->idx);
	    ind[1] = boxed_get(sigma_k);
	    coef[1] = 1;
	    ind[2] = st_v->sigma;
	    coef[2] = -1;
	    int lowb = GDD_vertex_delta_r(v, type) + period * GDD_max_distance_flowdep(u, v, type);
	    ilp_solver_add_row(scheduling_prob, namebuf, 2, ind, coef, ILP_SOLVER_LO, lowb);
	  }
	  SCEDA_hashset_iterator_cleanup(&consumers);
	}
	SCEDA_hashset_delete(cons_u);
      }
      SCEDA_hashmap_iterator_cleanup(&sigma_ks);
    }
    SCEDA_hashmap_iterator_cleanup(&vertices);
  }

#ifdef DEBUG
  {
    char filename[128];

    snprintf(filename, 128, "scheduling.problem.%d", period);
    ilp_solver_write_lp(scheduling_prob, filename);
  }
#endif

  if(ilp_solver_solve(scheduling_prob)) {
#ifdef DEBUG
    {
      char filename[128];
      snprintf(filename, 128, "scheduling.problem.%d.sol", period);
      
      FILE *stream = fopen(filename, "w+");

      SCEDA_VerticesIterator vertices;
      SCEDA_vertices_iterator_init(g, &vertices);
      while(SCEDA_vertices_iterator_has_next(&vertices)) {
	SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);
	State *st_v = SCEDA_hashmap_get(states, v);
	fprintf(stream,"sigma(%d) = %d\n", st_v->idx, ilp_solver_get_col_value(scheduling_prob, st_v->sigma));

	SCEDA_HashMapIterator sigma_ks;
	SCEDA_hashmap_iterator_init(st_v->sigma_k_map, &sigma_ks);
	while(SCEDA_hashmap_iterator_has_next(&sigma_ks)) {
	  char *type;
	  boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&sigma_ks, &type);
	  fprintf(stream,"sigma(%d,%s) = %d\n", st_v->idx, type, ilp_solver_get_col_value(scheduling_prob, boxed_get(sigma_k)));
	}
	SCEDA_hashmap_iterator_cleanup(&sigma_ks);
      }
      SCEDA_vertices_iterator_cleanup(&vertices);

      double tmp = ilp_solver_get_obj_value(scheduling_prob);
      fprintf(stream,"objective function = %.20g -> %d\n", tmp, (int)tmp); 

      fclose(stream);
    }
#endif

#ifdef PROBEOBJFUN
    obj_fun = rint(ilp_solver_get_obj_value(scheduling_prob));
#endif

    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }

      SCEDA_HashSet *vals = GDD_get_values_of_type(g, type);
      SCEDA_List *reuse_edges = SCEDA_list_create((SCEDA_delete_fun)SIRA_reuse_edge_delete);

      SIRA_solver_assignment_problem(scheduling_prob, vals, type, ual_flag, period, states, reuse_edges);

      SCEDA_hashmap_put(reuse_decisions, safe_strdup(type), reuse_edges, NULL);

      SCEDA_hashset_delete(vals);
    }
    SIRA_weight_map_iterator_cleanup(&weights);

    result = 0; // all is OK
  }

  ilp_solver_delete(scheduling_prob);

  SCEDA_hashmap_delete(states);

  ilp_solver_cleanup();

  return result;
}

#else

/** Solve the assignment problem

    @param[in] type = type t
    @param[in] killers = map values v of V_{R,t} to their killer k_v (of type t)
    @param[in] sigma = map vertices to their sigma
    @param[in] ual_flag = true if UAL semantics
    @param[in] period = scheduling period
    @param[in] reuse_edges = list of reuse edges (modified)

    return the cost of the minimal assignment
 */
static int SIRA_solver_assignment_problem_mcf(const char *type, SCEDA_HashMap *killers, SCEDA_HashMap *sigma,
					 int ual_flag, int period, 
					 SCEDA_List *reuse_edges) {
  int n = SCEDA_hashmap_size(killers);
  SCEDA_Vertex *values[n];

  {
    int i = 0;
    SCEDA_HashMapIterator vs;
    SCEDA_hashmap_iterator_init(killers, &vs);
    while(SCEDA_hashmap_iterator_has_next(&vs)) {
      SCEDA_Vertex *v;
      /* SCEDA_Vertex *kv = */ SCEDA_hashmap_iterator_next(&vs, &v);
      values[i++] = v;
    }
    SCEDA_hashmap_iterator_cleanup(&vs);
  }

  int **assignment_matrix = matrix_create(n);

  {
    int i;
    for(i = 0; i < n; i++) {
      SCEDA_Vertex *vi = values[i];
      SCEDA_Vertex *kvi = SCEDA_hashmap_get(killers, vi);

      int sigma_ki = *((int *)SCEDA_hashmap_get(sigma, kvi));

      int j;
      for(j = 0; j < n; j++) {
	SCEDA_Vertex *vj = values[j];
	int sigma_j = *((int *)SCEDA_hashmap_get(sigma, vj));
	
	int mu_ij = sigma_ki - sigma_j;
	if(ual_flag) {
	  mu_ij++;
	} else {
	  mu_ij -= GDD_vertex_delta_w(vj, type);
	}
	
	mu_ij += period - 1;
	if(mu_ij < 0) {
	  mu_ij = 0;
	}
	mu_ij /= period;
	
	assignment_matrix[i][j] = mu_ij;
      }
    }
  }

  hungarian_problem_t assignment_problem;

  int size = hungarian_init(&assignment_problem, assignment_matrix, n, n, HUNGARIAN_MODE_MINIMIZE_COST);
  assert(size == n);

  hungarian_solve(&assignment_problem);   
  
  int cost = 0;
  {
    int i;
    for(i = 0; i < n; i++) {
      int j;
      for(j = 0; j < n; j++) {
	if(assignment_problem.assignment[i][j] == HUNGARIAN_ASSIGNED) {
	  int mu_ij = assignment_matrix[i][j];
	  cost += mu_ij;
	  SIRA_ReuseEdge *e = SIRA_reuse_edge_create(values[i], values[j], mu_ij);
	  safe_call(SCEDA_list_add(reuse_edges, e));
	}
      }
    }
  }

  hungarian_free(&assignment_problem);

  matrix_delete(n, assignment_matrix);

  return cost;
}

/** Deduce sigma values and check optimality of flow 

    @param[in] g = DDG 
    @param[in] cost = cost of edges 
    @param[in] flow = min cost flow 

    @return a map giving for each vertex its sigma or NULL if the flow
    is not optimal */
static SCEDA_HashMap *SIRA_solver_deduce_sigma(SCEDA_Graph *g, SCEDA_HashMap *cost, SCEDA_HashMap *flow) {
  SCEDA_HashMap *dist = SCEDA_vertex_map_create((SCEDA_delete_fun)boxed_delete);

  int n = SCEDA_graph_vcount(g);

  SCEDA_HashSet *in_queue = SCEDA_vertex_set_create();
  SCEDA_Queue *queue = SCEDA_queue_create(NULL);
  
  /* Initially, each node is at distance 0 from a virtual source */
  {
    SCEDA_VerticesIterator vertices;
    SCEDA_vertices_iterator_init(g, &vertices);
    while(SCEDA_vertices_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);
      safe_call(SCEDA_hashmap_put(dist, v, boxed_create(int, 0), NULL));
      safe_call(SCEDA_queue_enqueue(queue, v));
      safe_call(SCEDA_hashset_add(in_queue, v));  
    }
    SCEDA_vertices_iterator_cleanup(&vertices);
  }

  safe_call(SCEDA_queue_enqueue(queue, NULL));

  /* This is an adaptation of Bellman-Ford algorithm, that works on the residual graph */
  int i = 0;
  while((i < n) && (!SCEDA_queue_is_empty(queue))) {
    SCEDA_Vertex *u;
    safe_call(SCEDA_queue_dequeue(queue, (void **)&u));
    if(u == NULL) {
      i++;
      if(i < n) {
	safe_call(SCEDA_queue_enqueue(queue, NULL));
      }
      continue;
    }
    safe_call(SCEDA_hashset_remove(in_queue, (void **)&u));
    
    boxed(int) dist_u = SCEDA_hashmap_get(dist, u);
   
    /* we first iterate over out edges with positive residual capacity */
    SCEDA_OutEdgesIterator out_edges;
    SCEDA_out_edges_iterator_init(u, &out_edges);
    while(SCEDA_out_edges_iterator_has_next(&out_edges)) {
      SCEDA_Edge *e = SCEDA_out_edges_iterator_next(&out_edges);
      /** every edge has unbounded capacity, so is in the residual graph */
      SCEDA_Vertex *v = SCEDA_edge_target(e);
      boxed(int) dist_v = SCEDA_hashmap_get(dist, v);
      int ce = boxed_get(((boxed(int))SCEDA_hashmap_get(cost, e)));
      if(boxed_get(dist_u) + ce < boxed_get(dist_v)) {
	boxed_set(dist_v, boxed_get(dist_u) + ce);
	if(!SCEDA_hashset_contains(in_queue, v)) {
	  safe_call(SCEDA_queue_enqueue(queue, v));
	  safe_call(SCEDA_hashset_add(in_queue, v));
	}
      }
    }
    SCEDA_out_edges_iterator_cleanup(&out_edges);
    
    /* then we iterate over in edges with positive residual capacity (flow) */
    SCEDA_InEdgesIterator in_edges;
    SCEDA_in_edges_iterator_init(u, &in_edges);
    while(SCEDA_in_edges_iterator_has_next(&in_edges)) {
      SCEDA_Edge *e = SCEDA_in_edges_iterator_next(&in_edges);
      boxed(int) fe = SCEDA_hashmap_get(flow, e);
      if(boxed_get(fe) <= 0) {
	continue;
      }
      SCEDA_Vertex *v = SCEDA_edge_source(e);
      boxed(int) dist_v = SCEDA_hashmap_get(dist, v);
      int ce = -boxed_get(((boxed(int))SCEDA_hashmap_get(cost, e)));
      if(boxed_get(dist_u) + ce < boxed_get(dist_v)) {
	boxed_set(dist_v, boxed_get(dist_u) + ce);
	if(!SCEDA_hashset_contains(in_queue, v)) {
	  safe_call(SCEDA_queue_enqueue(queue, v));
	  safe_call(SCEDA_hashset_add(in_queue, v));
	}
      }
    }
    SCEDA_in_edges_iterator_cleanup(&in_edges);
  }

  SCEDA_hashset_delete(in_queue);

  if(!SCEDA_queue_is_empty(queue)) {
    SCEDA_queue_delete(queue);
    SCEDA_hashmap_delete(dist);
    return NULL;
  }

  SCEDA_queue_delete(queue);

  {
    SCEDA_HashMapIterator vertices;
    SCEDA_hashmap_iterator_init(dist, &vertices);
    while(SCEDA_hashmap_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v;
      boxed(int) dist_v = SCEDA_hashmap_iterator_next(&vertices, &v);
      boxed_set(dist_v, -boxed_get(dist_v));
    }
    SCEDA_hashmap_iterator_cleanup(&vertices);
  }

  return dist;
}

/** Siralina solver (min cost flow implementation) 

    @param[in] loop = Siraloop DDG
    @param[in] settings = solver settings
    @param[in] period = scheduling period (II)
    @param[in] reuse_decisions = map (type -> reuse edges list) 

    return 0 in case of success, -1 otherwise */
static int SIRA_solver_fixed_period_mcf(SIRA_Loop *loop,
				   SIRA_Param *settings,
				   SIRA_WeightMap *weight_map,
				   int period,
				   SCEDA_HashMap *reuse_decisions) {
  SCEDA_Graph *g = SIRA_loop_ddg(loop);

  int ual_flag = settings->ual_flag;

  int result = -1; // error

  /* bookkeep added vertices/edges */
  SCEDA_List *Vk = SCEDA_list_create(NULL); 
  SCEDA_List *Ek = SCEDA_list_create(NULL);

  /* For each type t, a map v -> kv, where v is a value of type t and kv its killer */
  SCEDA_HashMap *killers_map = SCEDA_hashmap_create(NULL, (SCEDA_delete_fun)SCEDA_hashmap_delete, (SCEDA_compare_fun)string_match, (SCEDA_hash_fun)string_hash);

  SCEDA_HashMap *supply = SCEDA_vertex_map_create((SCEDA_delete_fun)boxed_delete);
  SCEDA_HashMap *cost = SCEDA_edge_map_create((SCEDA_delete_fun)boxed_delete);
  int bound = 0;

  /* initialise killers_map */
  {
    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }
      SCEDA_HashMap *killers = SCEDA_vertex_map_create(NULL);
      safe_call(SCEDA_hashmap_put(killers_map, type, killers, NULL));
    }
    SIRA_weight_map_iterator_cleanup(&weights);
  }

  /* We tweak the C type system */
  /* we first fill killers_map as follows: */
  /* For each type t, a map v -> cons(v, t), where v is a value of type t */
  {
    SCEDA_VerticesIterator vertices;
    SCEDA_vertices_iterator_init(g, &vertices);
    while(SCEDA_vertices_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);

      int v_supply = 0;

      {
	SIRA_WeightMapIterator weights;
	SIRA_weight_map_iterator_init(weight_map, &weights);
	while(SIRA_weight_map_iterator_has_next(&weights)) {
	  char *type;
	  int w = SIRA_weight_map_iterator_next(&weights, &type);
	  if(w < 0) {
	    continue;
	  }
	  if(GDD_vertex_is_value_of_type(v, type)) {
	    SCEDA_HashMap *killers = SCEDA_hashmap_get(killers_map, type);
	    SCEDA_HashSet *cons_v = GDD_consumers_of_type(v, type);
	    safe_call(SCEDA_hashmap_put(killers, v, cons_v, NULL));
	    v_supply += w;
	  }
	}
	SIRA_weight_map_iterator_cleanup(&weights);
      }

      safe_call(SCEDA_hashmap_put(supply, v, boxed_create(int, v_supply), NULL));
      bound += 2*v_supply;
    }
    SCEDA_vertices_iterator_cleanup(&vertices);
  }

  {
    SCEDA_EdgesIterator edges;
    SCEDA_edges_iterator_init(g, &edges);
    while(SCEDA_edges_iterator_has_next(&edges)) {
      SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);
      int c = GDD_edge_latency(e) - period * GDD_edge_distance(e);
      safe_call(SCEDA_hashmap_put(cost, e, boxed_create(int, -c), NULL));
    }
    SCEDA_edges_iterator_cleanup(&edges);
  }

  {
    SCEDA_HashMapIterator kmap;
    SCEDA_hashmap_iterator_init(killers_map, &kmap);
    while(SCEDA_hashmap_iterator_has_next(&kmap)) {
      char *type;
      SCEDA_HashMap *killers = SCEDA_hashmap_iterator_next(&kmap, &type);

      int w = SIRA_weight_map_get_weight_of_type(weight_map, type);

      SCEDA_HashMapIterator consumers;
      SCEDA_hashmap_iterator_init(killers, &consumers);
      while(SCEDA_hashmap_iterator_has_next(&consumers)) {
	SCEDA_Vertex *u;
	SCEDA_HashSet *cons_u = SCEDA_hashmap_iterator_next(&consumers, &u);

	SCEDA_Vertex *uk = SCEDA_graph_add_vertex(g, NULL);
	safe_call(SCEDA_list_add(Vk, uk));

	safe_call(SCEDA_hashmap_put(supply, uk, boxed_create(int, -w), NULL));

	/* the following line is correct because it does not modify
	   the structure of the container */
	assert(SCEDA_hashmap_put(killers, u, uk, NULL) == 1); 

	if(SCEDA_hashset_is_empty(cons_u)) {
	  SCEDA_Edge *ek = SCEDA_graph_add_edge(g, u, uk, NULL);
	  safe_call(SCEDA_list_add(Ek, ek));
	  int c = 1;
	  safe_call(SCEDA_hashmap_put(cost, ek, boxed_create(int, -c), NULL));
	} else {
	  SCEDA_HashSetIterator cons;
	  SCEDA_hashset_iterator_init(cons_u, &cons);
	  while(SCEDA_hashset_iterator_has_next(&cons)) {
	    SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&cons);
	    SCEDA_Edge *ek = SCEDA_graph_add_edge(g, v, uk, NULL);
	    safe_call(SCEDA_list_add(Ek, ek));
	    
	    /* the following line is correct because neither u nor v is
	       a killer node and thus, the edges from u to v are not fake */
	    int c = GDD_vertex_delta_r(v, type) + period * GDD_max_distance_flowdep(u, v, type); 
	    safe_call(SCEDA_hashmap_put(cost, ek, boxed_create(int, -c), NULL));
	  }
	  SCEDA_hashset_iterator_cleanup(&cons);
	}

	SCEDA_hashset_delete(cons_u);
      }
      SCEDA_hashmap_iterator_cleanup(&consumers);
    }
    SCEDA_hashmap_iterator_cleanup(&kmap);
  }

  SCEDA_HashMap *flow = SCEDA_graph_min_cost_flow(g, NULL, NULL, NULL, &bound, NULL, supply, NULL, cost);
  if(flow != NULL) {
    SCEDA_HashMap *sigma = SIRA_solver_deduce_sigma(g, cost, flow);
    /** We can forget the flow */
    SCEDA_hashmap_delete(flow); 
    if(sigma != NULL) {
      SCEDA_HashMapIterator killers;
      SCEDA_hashmap_iterator_init(killers_map, &killers);
      while(SCEDA_hashmap_iterator_has_next(&killers)) {
	char *type;
	SCEDA_HashMap *kmap = SCEDA_hashmap_iterator_next(&killers, &type);
	
	SCEDA_List *reuse_edges = SCEDA_list_create((SCEDA_delete_fun)SIRA_reuse_edge_delete);
	
	SIRA_solver_assignment_problem_mcf(type, kmap, sigma, ual_flag, period, reuse_edges);
	
	SCEDA_hashmap_put(reuse_decisions, safe_strdup(type), reuse_edges, NULL);
      }
      SCEDA_hashmap_iterator_cleanup(&killers);
      
      SCEDA_hashmap_delete(sigma);
      
      result = 0; // a solution was found
    }
  }

  {
    SCEDA_ListIterator edges;
    SCEDA_list_iterator_init(Ek, &edges);
    while(SCEDA_list_iterator_has_next(&edges)) {
      SCEDA_Edge *ek = SCEDA_list_iterator_next(&edges);

      void *data;
      SCEDA_graph_remove_edge(g, ek, &data);
    }
    SCEDA_list_iterator_cleanup(&edges);

    SCEDA_list_delete(Ek);
  }

  {
    SCEDA_ListIterator vertices;
    SCEDA_list_iterator_init(Vk, &vertices);
    while(SCEDA_list_iterator_has_next(&vertices)) {
      SCEDA_Vertex *vk = SCEDA_list_iterator_next(&vertices);

      void *data;
      SCEDA_graph_remove_vertex(g, vk, &data);
    }
    SCEDA_list_iterator_cleanup(&vertices);

    SCEDA_list_delete(Vk);
  }

  SCEDA_hashmap_delete(killers_map);

  SCEDA_hashmap_delete(supply);
  SCEDA_hashmap_delete(cost);

  return result;
}

#endif

/** Check whether the solution satisfies the register constraints.

    @param[in] regconstraints_map = register constraints
    @param[in] reuse_decisions = map (type -> reuse edges list)

    @return TRUE iff constraints are satisfied */
static int SIRA_solver_is_solution(SIRA_RegConstraintsMap *regconstraints_map, SIRA_WeightMap *weight_map, SCEDA_HashMap *reuse_decisions) {
  int is_solution = TRUE;
  //
  SCEDA_HashMapIterator decisions;
  SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
  while(SCEDA_hashmap_iterator_has_next(&decisions)) {
    char *type;
    SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
    int w = SIRA_weight_map_get_weight_of_type(weight_map, type);
    if(w <= 0) {
      continue;
    }
    int cost = SIRA_reuse_decisions_get_cost(reuse_edges);
    int max = SIRA_regconstraints_map_get_constraint_of_type(regconstraints_map, type);
    if(cost > max) {
      is_solution = FALSE;
      break;
    }
  }
  SCEDA_hashmap_iterator_cleanup(&decisions);
  //
  return is_solution;
}

/** Check whether the solution is valid, i.e. if the resulting DDG is
    lexicographic positive.

    @param[in] loop = SiraLoop DDG
    @param[in] settings = solver settings
    @param[in] reuse_decisions = map (type -> reuse edges list)

    @return TRUE iff resulting DDG is lexicographic positive 
    note that it is trivially TRUE when UAL semantics is enabled */
static int SIRA_solver_solution_is_lexicographic_positive(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_HashMap *reuse_decisions) {
  int result = TRUE;
  if(SIRA_param_get_ual_flag(settings)) {
    return result;
  }

  /* check whether the modified DDG is lexicographic positive */
  /* apply the reuse decisions */
  SCEDA_List *all_antidep_edges = SCEDA_list_create((SCEDA_delete_fun)SCEDA_list_delete);
  {
    SCEDA_HashMapIterator decisions;
    SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
    while(SCEDA_hashmap_iterator_has_next(&decisions)) {
      char *type;
      SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
      SCEDA_List *antidep_edges = SIRA_reuse_apply_decisions(SIRA_loop_ddg(loop), type, reuse_edges, SIRA_param_get_ual_flag(settings));
      safe_call(SCEDA_list_add(all_antidep_edges, antidep_edges));
    }
    SCEDA_hashmap_iterator_cleanup(&decisions);
  }

  /* check lexicographic positiveness */
  SCEDA_List *cycle;
  if(!GDD_is_lexicographic_positive(SIRA_loop_ddg(loop), &cycle)) {
    SCEDA_list_delete(cycle);
    result = FALSE;
  }

  /* unapply reuse decisions */
  {
    SCEDA_ListIterator ae;
    SCEDA_list_iterator_init(all_antidep_edges, &ae);
    while(SCEDA_list_iterator_has_next(&ae)) {
      SCEDA_List *antidep_edges = SCEDA_list_iterator_next(&ae);
      GDD_delete_edges(SIRA_loop_ddg(loop), antidep_edges);
    }
    SCEDA_list_iterator_cleanup(&ae);
  }
  SCEDA_list_delete(all_antidep_edges);
  
  return result;
}

/** Apply siralina heuristic for every II between IImin and IImax and
    print the results (used for statistics).

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) */
void SIRA_solver_stats(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax) {
  if(IImin < 0) {
    IImin = SIRA_loop_IImin(loop);
  }
  if(IImax < 0) {
    IImax = SIRA_loop_IImax(loop);
  }

  SCEDA_HashMap *reuse_decisions = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  int II;
  for(II = IImin; II <= IImax; II++) {
    if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
      fprintf(stdout,"II= %d\tL= %d\t", II, SIRA_loop_get_L(loop, II));

      SCEDA_HashMapIterator decisions;
      SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
      while(SCEDA_hashmap_iterator_has_next(&decisions)) {
	char *type;
	SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
	int cost = SIRA_reuse_decisions_get_cost(reuse_edges);
	fprintf(stdout, "R_%s= %d\t", type, cost);
      }
      SCEDA_hashmap_iterator_cleanup(&decisions);

#ifdef PROBEOBJFUN
      fprintf(stdout,"obj= %d\t", obj_fun);
#endif

      fprintf(stdout,"\n");
    }
    SCEDA_hashmap_clear(reuse_decisions);
  }
  
  SCEDA_hashmap_delete(reuse_decisions);
}

/** Apply siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] II = II (if negative, fails)
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return 0 if a solution was found, otherwise -1 */
static int SIRA_solver_single_internal(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int II, SCEDA_HashMap **solution) {
  if(II < 0) {
    return -1;
  }

  SCEDA_HashMap *reuse_decisions = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
    *solution = reuse_decisions;
    return 0;
  } else {
    SCEDA_hashmap_delete(reuse_decisions);
    return -1;
  }
}

/** Apply siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] II = II (if negative, fails)
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return 0 if a solution was found, otherwise -1 */
int SIRA_solver_single(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int II, SCEDA_HashMap **solution) {
  if(SIRA_solver_single_internal(loop, settings, weight_map, II, solution) == 0) {
    /* Do we have to check the result ? */
    if(!SIRA_param_get_check_flag(settings)) {
      return 0;
    }
    if(SIRA_param_get_ual_flag(settings)) {
      /* if UAL semantics then no need to check solution */
      return 0;
    } else {
      if(SIRA_solver_solution_is_lexicographic_positive(loop, settings, *solution)) {
	return 0;
      } 
      /* the solution was not lexicographic positive */
      /* retry with UAL semantics */
      SCEDA_hashmap_delete(*solution);
      SIRA_param_set_ual_flag(settings, TRUE);
      return SIRA_solver_single_internal(loop, settings, weight_map, II, solution);
    }
  }
  return -1;
}

/** Apply siralina heuristic for every II between IImin and IImax and
    stops when a solution is found (ie satisfies the register
    constraints).

    Before beginning the search procedure, a simple heuristic is
    implemented to determine whether there is a solution to the
    problem.

    @param[in] loop = siraloop
    @param[in] regconstraints_map = register contraints
    @param[in] settings = solver settings
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return II >= 0 if a solution was found, otherwise -1 */
static int SIRA_solver_search_internal(SIRA_Loop *loop, SIRA_RegConstraintsMap *regconstraints_map, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax, SCEDA_HashMap **solution) {
  int found = FALSE;

  if(IImin < 0) {
    IImin = SIRA_loop_IImin(loop);
  }
  if(IImax < 0) {
    IImax = SIRA_loop_IImax(loop);
  }
  if(IImax < 0) {
    return -1;
  }
  if(IImax < IImin) {
    return -1;
  }

  SCEDA_HashMap *reuse_decisions = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  int II = IImin;
  SCEDA_hashmap_clear(reuse_decisions);
  if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
    found = SIRA_solver_is_solution(regconstraints_map, weight_map, reuse_decisions);
  }

  if(!found) {
    II = IImax;
    SCEDA_hashmap_clear(reuse_decisions);
    if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
      found = SIRA_solver_is_solution(regconstraints_map, weight_map, reuse_decisions);
    }

    if(found) {
      found = FALSE;
      for(II = IImin+1; II <= IImax; II++) {
	SCEDA_hashmap_clear(reuse_decisions);
	if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
	  found = SIRA_solver_is_solution(regconstraints_map, weight_map, reuse_decisions);
	}
	if(found) {
	  break;
	}
      }
    }
  }

  if(found) {
    *solution = reuse_decisions;
    return II;
  } else {
    SCEDA_hashmap_delete(reuse_decisions);
    return -1;
  }
}

/** Apply siralina heuristic for every II between IImin and IImax and
    stops when a solution is found (ie satisfies the register
    constraints).

    Before beginning the search procedure, a simple heuristic is
    implemented to determine whether there is a solution to the
    problem.

    @param[in] loop = siraloop
    @param[in] regconstraints_map = register contraints
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return II >= 0 if a solution was found, otherwise -1 */
int SIRA_solver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *regconstraints_map, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax, SCEDA_HashMap **solution) {
  int II = SIRA_solver_search_internal(loop, regconstraints_map, settings, weight_map, IImin, IImax, solution);
  if(II < 0) {
    return II;
  }
  /** Do we have to check the result ? */
  if(!SIRA_param_get_check_flag(settings)) {
    return II;
  }
  /** If UAL semantics, no need to check */
  if(SIRA_param_get_ual_flag(settings)) {
    return II;
  }
  if(SIRA_solver_solution_is_lexicographic_positive(loop, settings, *solution)) {
    return II;
  } 
  /* the solution was not lexicographic positive */
  /* retry with UAL semantics */
  SCEDA_hashmap_delete(*solution);
  SIRA_param_set_ual_flag(settings, TRUE);
  return SIRA_solver_search_internal(loop, regconstraints_map, settings, weight_map, IImin, IImax, solution);
}

