/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
/** \file weightmap.h
    \brief Specifying weights of types */
#ifndef __SIRA_WEIGHTMAP_H
#define __SIRA_WEIGHTMAP_H

#include <SCEDA/hashmap.h>

/** Type of the weight map 

    \hideinitializer */
typedef SCEDA_HashMap SIRA_WeightMap;

/** Create an empty weight map.

    @return a new weight map
*/
SIRA_WeightMap *SIRA_weight_map_create();

/** Delete a weight map.

    @param[in] weight_map = weight map to delete 

    \hideinitializer */
#define SIRA_weight_map_delete SCEDA_hashmap_delete

/** Set weight of given type.

    @param[in] weight_map = weight map
    @param[in] type = type
    @param[in] weight = new weight

    @return 0 in case of success, <>0 otherwise */
int SIRA_weight_map_set_weight_of_type(SIRA_WeightMap *weight_map, const char *type, int weight);

/** Return weight of given type in the weight map.

    @param[in] weight_map = weight map
    @param[in] type = wanted type

    @return weight of type or -1 if undefined */
int SIRA_weight_map_get_weight_of_type(SIRA_WeightMap *weight_map, const char *type);

/** Is the weight map empty?

    @param[in] weight_map = weight map

    @return TRUE iff the given weight map is empty 

    \hideinitializer */
#define SIRA_weight_map_is_empty SCEDA_hashmap_is_empty

/** Clear the weight map.

    @param[in] weight_map = weight map to clear 

    \hideinitializer */
#define SIRA_weight_map_clear SCEDA_hashmap_clear

/** Type of the weight map iterators.

    \hideinitializer */
typedef SCEDA_HashMapIterator SIRA_WeightMapIterator;

/** Initialise a weight map iterator 

    \hideinitializer */
#define SIRA_weight_map_iterator_init SCEDA_hashmap_iterator_init

/** Is there a next element in a weight map iterator?

    \hideinitializer */
#define SIRA_weight_map_iterator_has_next SCEDA_hashmap_iterator_has_next

/** Return the next element in the weight map iterator

    @param[in] iter = iterator
    @param[out] type = type

    @return the weight of the type
    (will fail if iterator is empty) */
int SIRA_weight_map_iterator_next(SIRA_WeightMapIterator *iter, char **type);

/** Cleanup a weight map iterator 

    \hideinitializer */
#define SIRA_weight_map_iterator_cleanup SCEDA_hashmap_iterator_cleanup

#endif

/** \page weight_map Setting weights of types

    The weight of each type is given by a SIRA_WeightMap structure.

    A SIRA_WeightMap is created thanks to SIRA_weight_map_create
    function and deleted thanks to SIRA_weight_map_delete function.

    We advise to access the SIRA_WeightMap structures exclusively
    through the functions provided by the API.
    
    \section weight_map_api API

    \code
    SIRA_WeightMap *SIRA_weight_map_create();
    \endcode

    Create an empty weight map.

    \code
    void SIRA_weight_map_delete(SIRA_WeightMap *weight_map);
    \endcode

    Delete a weight map.

    \code
    int SIRA_weight_map_set_weight_of_type(SIRA_WeightMap *weight_map, const char *type, int weight);
    \endcode

    Set weight of given type in the weight map. Fails when the weight is negative or already defined.

    \code
    int SIRA_weight_map_get_weight_of_type(SIRA_WeightMap *weight_map, const char *type);
    \endcode

    Get weight of given type in the weight map. Return -1 when weight of given type is undefined.

    \code
    void SIRA_weight_map_clear(SIRA_WeightMap *weight_map);
    \endcode

    Empty a weight map.

    \code
    int SIRA_weight_map_is_empty(SIRA_WeightMap *weight_map);
    \endcode

    Test the emptiness of a weight map.

    \code
    void SIRA_weight_map_iterator_init(SIRA_WeightMap *weight_map, SIRA_WeightMapIterator *iter);
    int SIRA_weight_map_iterator_has_next(SIRA_WeightMapIterator *iter);
    int SIRA_weight_map_iterator_next(SIRA_WeightMapIterator *iter, char **type);
    void SIRA_weight_map_iterator_cleanup(SIRA_WeightMap *weight_map);
    \endcode

    Iterator over weight maps.
*/
